from .base_node import Node
from .crypt import Crypt, universal_decode, universal_encode

from tempfile import TemporaryFile
from typing import Any
from io import FileIO, BytesIO, IOBase, BufferedRandom
import os, base64, hashlib
import math
import json
from logging import exception


class AdvancedNode(
    Node
):  # Node that implements high-level processes like chunking and multiple target selection
    def __init__(
        self,
        name: str,
        network_id: str,
        crypt: Crypt = None,
        network_key: bytes | None = None,
        server_port: int = 3333,
        advertise_port: int = 3334,
        advertise_interval: float | int = 2.5,
        bind_ip: str = "localhost",
        chunk_size: int = 16384,
        functions: dict = {},
    ):
        if not all(
            [
                i.lower()
                in "1234567890qwertyuiopasdfghjklzxcvbnm!@#$%^&*()[]{}:,<.>/-_+`~"
                for i in name
            ]
        ):
            raise ValueError(
                'All name characters must be in "1234567890qwertyuiopasdfghjklzxcvbnm!@#$%^&*()[]{}:,<.>/-_+`~"'
            )
        super().__init__(
            name,
            network_id,
            self.handle,
            crypt,
            network_key,
            server_port,
            advertise_port,
            advertise_interval,
            bind_ip,
        )
        self.current_chunked = {}
        self.chunk_size = chunk_size
        self.functions = functions

    def register(self, command: str):
        """
        Decorated functions should take 3 arguments: The Node instance, the originator name, and a readable IOStream.
        They should return a readable IOStream or any object. IOStreams will be chunked and returned to the originator that way,
        while other objects will be returned as a raw universal stream
        """

        def register_decorator(func):
            self.functions[command] = func
            return func

        return register_decorator

    def _parse_args(self, args: str):
        return args.split("?")[0], {
            x.split("=")[0]: x.split("=")[1] for x in args.split("?")[1].split("&")
        }

    def _send_chunked(
        self, target: str, command: str, data_or_stream: bytes | FileIO
    ) -> IOBase:
        try:
            if type(data_or_stream) == bytes:
                data_size = len(data_or_stream)
                data_or_stream = BytesIO(data_or_stream)
            else:
                data_or_stream.seek(0, 2)
                data_size = data_or_stream.tell()
                data_or_stream.seek(0)

            stream_id = base64.urlsafe_b64encode(os.urandom(8))

            super().send(
                target,
                f"CHUNKED?pt=initiator&ln={math.ceil(data_size/self.chunk_size)}&id={stream_id}&og={self.name}&fn={command}",
            )
            chunk_index = 0
            shas = []
            while True:
                last_pos = data_or_stream.tell()
                new_data = data_or_stream.read(self.chunk_size)
                if not new_data:
                    break
                new_data = base64.urlsafe_b64encode(new_data)
                sha = hashlib.sha256(new_data).hexdigest()
                new_data = new_data.decode("utf-8").replace("=", "~")
                chunk_result = super().send(
                    target,
                    f"CHUNKED?pt=part&id={stream_id}&in={chunk_index}&dt={new_data}&ck={sha}&og={self.name}",
                )
                if not "?" in chunk_result:
                    raise ValueError("Invalid message")
                cmd, args = self._parse_args(chunk_result)
                if cmd == "RESPONSE" and args["rs"] == "success":
                    chunk_index += 1
                    shas.append(sha[:6])
                    continue
                data_or_stream.seek(last_pos)
            result = super().send(
                target,
                f"CHUNKED?pt=end&og={self.name}&ck={hashlib.sha256(''.join(shas).encode('utf-8')).hexdigest()}&id={stream_id}",
            )
            if not "?" in result:
                raise ValueError("Invalid message")
            cmd, args = self._parse_args(result)
            if cmd == "RESPONSE" and args["rs"] == "success":
                return BytesIO(
                    base64.urlsafe_b64decode(args["dt"].replace("~", "=").encode("utf-8"))
                )
            elif cmd == "CHUNKED_RESPONSE" and args["rs"] == "success":
                chunk = {
                    "num_packets": int(args["ln"]),
                    "originator": args["og"],
                    "data_storage": TemporaryFile(),
                    "id": args["id"],
                }
                c = 0
                shas = []
                while c <= chunk["num_packets"]:
                    chunk_result = super().send(
                        target, f"CHUNKED_REQUEST?pt=part&id={chunk['id']}&in={c}"
                    )
                    cmd, args = self._parse_args(chunk_result)
                    if cmd == "CHUNKED_RESPONSE":
                        if args["pt"] == "part":
                            checksum = hashlib.sha256(
                                args["dt"].replace("~", "=").encode("utf-8")
                            ).hexdigest()
                            if args["ck"] == checksum:
                                chunk["data_storage"].write(
                                    base64.urlsafe_b64decode(
                                        args["dt"].replace("~", "=").encode("utf-8")
                                    )
                                )
                                shas.append(checksum[:6])
                                c += 1
                                continue
                        elif args["pt"] == "end":
                            if (
                                args["ck"]
                                == hashlib.sha256("".join(shas).encode("utf-8")).hexdigest()
                            ):
                                chunk["data_storage"].seek(0)
                                return chunk["data_storage"]
                    raise ValueError(f"Error packet: {cmd}:{args}")
            raise ValueError("Invalid response")
        except:
            exception()
            return "$error"

    def handle(self, content: str):
        cmd, args = self._parse_args(content)
        if cmd == "CHUNKED":
            if args["pt"] == "initiator":
                if not args["fn"] in self.functions.keys():
                    return "ERROR?rs=unknown function"
                self.current_chunked[args["id"]] = {
                    "length": args["ln"],
                    "originator": args["og"],
                    "type": "request",
                    "storage": TemporaryFile(),
                    "hashes": [],
                    "command": args["fn"],
                }
                return "RESPONSE?rs=success"
            elif args["pt"] == "part":
                if args["id"] in self.current_chunked.keys():
                    checksum = hashlib.sha256(
                        args["dt"].replace("~", "=").encode("utf-8")
                    ).hexdigest()
                    if args["ck"] == checksum:
                        self.current_chunked[args["id"]]["hashes"].append(checksum[:6])
                        self.current_chunked[args["id"]]["storage"].write(
                            base64.urlsafe_b64decode(
                                args["dt"].replace("~", "=").encode("utf-8")
                            )
                        )
                        return "RESPONSE?rs=success"
                    else:
                        return "ERROR?rs=checksum error"
                else:
                    return "ERROR?rs=wrong id"
            else:
                if args["id"] in self.current_chunked.keys():
                    checksum = hashlib.sha256(
                        "".join(self.current_chunked[args["id"]]["hashes"]).encode(
                            "utf-8"
                        )
                    ).hexdigest()
                    if args["ck"] == checksum:
                        self.current_chunked[args["id"]]["storage"].seek(0)
                        result = self.functions[
                            self.current_chunked[args["id"]]["command"]
                        ](
                            self,
                            self.current_chunked[args["id"]]["originator"],
                            self.current_chunked[args["id"]]["storage"],
                        )
                    else:
                        return "ERROR?rs=checksum error"
                else:
                    return "ERROR?rs=wrong id"
        elif cmd == "CHUNKED_REQUEST":
            if args["id"] in self.current_chunked.keys():
                data = self.current_chunked[args["id"]]["storage"].read(self.chunk_size)
                if data:
                    data = base64.urlsafe_b64encode(data)
                    checksum = hashlib.sha256(data).hexdigest()
                    self.current_chunked[args["id"]]["hashes"].append(checksum[:6])
                    return f"CHUNKED_RESPONSE?pt=part&dt={data.decode('utf-8').replace('=', '~')}&ck={checksum}"
                else:
                    ret = f"CHUNKED_RESPONSE?pt=end&ck={hashlib.sha256(''.join(self.current_chunked[args['id']]['hashes']).encode('utf-8')).hexdigest()}"
                    del self.current_chunked[args["id"]]
                    return ret
            else:
                return "ERROR?rs=wrong id"
        elif cmd == "RAW":
            result = self.functions[args["fn"]](
                self,
                args["og"],
                BytesIO(
                    base64.urlsafe_b64decode(
                        args["dt"].replace("~", "=").encode("utf-8")
                    )
                ),
            )
        else:
            return "ERROR?rs=unknown command"

        if isinstance(result, IOBase) and result.seekable():
            result.seek(0, 2)
            _size = math.ceil(result.tell() / self.chunk_size)
            result.seek(0)
            stream_id = (
                base64.urlsafe_b64encode(os.urandom(8)).decode("utf-8").strip("=")
            )
            self.current_chunked[stream_id] = {
                "length": _size,
                "originator": self.name,
                "type": "response",
                "storage": result,
                "hashes": [],
            }
            return (
                f"CHUNKED_RESPONSE?rs=success&ln={_size}&og={self.name}&id={stream_id}"
            )
        else:
            return f"RESPONSE?rs=success&dt={universal_encode(result).decode('utf-8').replace('=', '~')}"

    def send(self, target: str | list[str], command: str, data: Any) -> Any:
        # Target selection
        if target == "*":
            targets = list(self.peers.keys())
        elif type(target) == list:
            targets = [t for t in target if t in self.peers.keys()]
        else:
            targets = [target]

        # Data encoding and preparation for chunking
        if isinstance(data, IOBase):
            to_send = data
            encoded = False
        else:
            enc_data = universal_encode(data)
            encoded = True
            if len(enc_data) > self.chunk_size:
                to_send = BytesIO(enc_data)
            else:
                to_send = enc_data

        if isinstance(to_send, IOBase):
            cmds = [command for c in targets]
            datas = [data for d in targets]
            results = map(self._send_chunked, targets, cmds, datas)
        else:
            datas = [
                f"RAW?fn={command}&og={self.name}&dt={enc_data.decode('utf-8').replace('=', '~')}"
                for t in targets
            ]
            results = [
                self._parse_args(i)[1]["dt"].replace("~", "=")
                for i in map(super().send, targets, datas)
            ]

        results = list(results)
        if len(results) == 1:
            return (
                universal_decode(
                    results[0] if isinstance(results[0], bytes) else results[0].read()
                )
                if encoded and results[0] != "$error"
                else results[0]
            )
        else:
            return (
                {
                    targets[i]: universal_decode(
                        results[i]
                        if isinstance(results[i], bytes) or isinstance(results[i], str)
                        else results[i].read()
                    )
                    if encoded and results[i] != "$error"
                    else results[i]
                    for i in range(len(targets))
                }
                if encoded
                else {targets[i]: results[i] for i in range(len(targets))}
            )

class Target:
    def __init__(self, node: AdvancedNode, target: str | list[str]) -> None:
        self.target = target
        self.origin = node
    
    def __getattr__(self, item):
        return CommandFunction(self, item)

class CommandFunction:
    def __init__(self, target: Target, func: str) -> None:
        self.origin = target
        self.function = func
    
    def __call__(self, *args: Any, **kwds: Any) -> Any:
        result = self.origin.origin.send(self.origin.target, "handle", BytesIO(json.dumps({
            "function": self.function,
            "args": args,
            "kwargs": kwds
        }).encode("utf-8")))
        if isinstance(result, BufferedRandom):
            result = json.loads(result.read().decode("utf-8"))
            if result["type"] == "success":
                return result["value"]
            else:
                raise ConnectionError(f"Failed to get result: error {result['description']}")
        else:
            ret = {}
            for k in result.keys():
                r = json.loads(result[k].read().decode("utf-8"))
                if r["type"] == "success":
                    ret[k] = r["value"]
                else:
                    ret[k] = f"Failed to get result: error {r['description']}"
            
            return ret


# Node that implements utility functions for RPC
class CommandNode(AdvancedNode):
    def __init__(
        self,
        name: str,
        network_id: str,
        crypt: Crypt = None,
        network_key: bytes | None = None,
        server_port: int = 3333,
        advertise_port: int = 3334,
        advertise_interval: float | int = 2.5,
        bind_ip: str = "localhost",
        chunk_size: int = 16384,
        functions: dict = {},
    ):
        self.registered_functions = functions
        self.registered_functions["get_peers"] = self._get_peers
        self.registered_functions["get_funcs"] = self._get_funcs
        super().__init__(
            name,
            network_id,
            crypt,
            network_key,
            server_port,
            advertise_port,
            advertise_interval,
            bind_ip,
            chunk_size,
            {"handle": self._handle},
        )
    
    def _handle(self, node: AdvancedNode, originator: str, data: BytesIO):
        try:
            args = json.loads(data.read().decode("utf-8"))
        except json.JSONDecodeError:
            return BytesIO(json.dumps({
                "type": "error",
                "description": "bad json data"
            }).encode("utf-8"))
        if not all([i in args.keys() for i in ["function", "args", "kwargs"]]):
            return BytesIO(json.dumps({
                "type": "error",
                "description": "not all required items (function, args, kwargs) are present"
            }).encode("utf-8"))
        if not args["function"] in self.registered_functions.keys():
            return BytesIO(json.dumps({
                "type": "error",
                "description": "function does not exist on node"
            }).encode("utf-8"))

        result = self.registered_functions[args["function"]](self, originator, *args["args"], **args["kwargs"])
        try:
            return BytesIO(json.dumps({
                "type": "success",
                "value": result
            }).encode("utf-8"))
        except:
            return BytesIO(json.dumps({
                "type": "error",
                "description": "bad return data"
            }).encode("utf-8"))
    
    def _get_peers(self, *args, **kwargs):
        return list(self.peers.keys())
    
    def _get_funcs(self, *args, **kwargs):
        return list(self.registered_functions.keys())
    
    def register(self, command: str):
        def register_decorator(func):
            self.registered_functions[command] = func
            return func

        return register_decorator

    def target(self, target: str | list[str]) -> Target:
        return Target(self, target)
        




"""
PNET - Local p2p protocol

(c) Dax Harris 2021
MIT License
"""

import rsa
import rsa.pkcs1
from cryptography.fernet import Fernet, InvalidToken
import base64
import pickle
import random
import typing
import socket
import socketserver
import time
from threading import Thread
from logging import warning, exception
import re
from .crypt import Crypt
        

class Node:
    def __init__(
        self,
        name: str,
        network_id: str,
        onmessage = print,
        crypt: Crypt = None,
        network_key: bytes | None = None,
        server_port: int = 3333,
        advertise_port: int = 3334,
        advertise_interval: float | int = 2.5,
        bind_ip: str = "localhost"
    ):
        """
        P2P Node
        :name Unique name within network - Should not use | or @
        :network_id ID of network to listen to - Should not use |
        :onmessage Callback to run when a message is recieved. Takes one argument
        :crypt Crypt object or None. If None, auto-generates object.
        :network_key Fernet encryption key of network. Leaving as None will disable network-level encryption
        :server_port Port to listen on
        :advertise_port Port to advertise on. Should be shared across a network
        :broadcast_interval Time between UDP advertisements
        :bind_ip IP to bind to.
        """
        
        self.name = name
        self.network_id = network_id
        if network_key:
            self.fernet = Fernet(network_key)
            self.net_encrypted = True
        else:
            self.net_encrypted = False
            self.fernet = None
        
        self.server_port = server_port
        self.advertise_port = advertise_port
        self.peers = {}
        self.ad_interval = advertise_interval
        self.running = False
        self.bind_ip = bind_ip
        self.callback = onmessage
        if crypt:
            self.crypt = crypt
        else:
            self.crypt = Crypt()
        self.pk = base64.urlsafe_b64encode(self.crypt.public.save_pkcs1()).decode("utf-8")
        self.sserver = None
    
    def _advertiser(self):
        advertiser = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        advertiser.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
        advertiser.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        advertiser.bind(("", 0))
        advertiser.settimeout(0.001)
        msgcontent = f"{self.name}@{socket.gethostbyname(socket.gethostname()) if self.bind_ip in ['', '0.0.0.0'] else self.bind_ip}:{self.server_port}"
        if self.net_encrypted:
            msgcontent = base64.urlsafe_b64encode(self.fernet.encrypt(msgcontent.encode("utf-8"))).decode("utf-8")
        broadcast_message = f"{self.network_id}|adv|{msgcontent}|{self.pk}|END\n".encode("utf-8")
        while self.running:
            advertiser.sendto(broadcast_message, ("<broadcast>", self.advertise_port))
            time.sleep(self.ad_interval)
        advertiser.close()
    
    def _proc_udp(self, data):
        data = data.strip()
        parts = data.decode("utf-8").split("|")
        if "brd" in parts:
            print(len(parts))
        if len(parts) != 5:
            return
        if parts[4] != "END":
            return
        if parts[0] != self.network_id:
            return

        _, ptype, content, pk, _ = parts

        # Check packet type
        if not ptype == "adv":
            return
        
        # Attempt to decrypt packet on network level
        if self.net_encrypted:
            try:
                content: bytes = self.fernet.decrypt(base64.urlsafe_b64decode(content.encode("utf-8")))
            except InvalidToken:
                return
        else:
            content: bytes = content.encode("utf-8")
        
        content = content.decode("utf-8")
        if "@" in content and ":" in content:
            name, addr = content.split("@")
            if name == self.name:
                return
            if not re.search("^((25[0-5]|2[0-4][0-9]|1[0-9][0-9]|[1-9]?[0-9])\.){3}(25[0-5]|2[0-4][0-9]|1[0-9][0-9]|[1-9]?[0-9]):[0-9]{1,5}$", addr) and not addr.startswith("localhost:"):
                return
            
            self.peers[name] = {
                "name": name,
                "addr": addr,
                "public_key": rsa.PublicKey.load_pkcs1(base64.urlsafe_b64decode(pk.encode("utf-8")))
            }

    def discover(self):
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM) #create UDP socket
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        s.bind(('', self.advertise_port))
        while self.running:
            datas = b""
            while True:
                _data, _ = s.recvfrom(1024)
                if not _data: break
                datas += _data
                if b"END\n" in datas:
                    packs = datas.split(b"\n")
                    datas = packs.pop()
                    [Thread(target=self._proc_udp, daemon=True, args=[i,]).start() for i in packs]
        s.close()
            
    
    def create_handler(self, *args):
        return NodeHandler(self, *args)
    
    def serve(self):
        """
        Runs node server
        """
        print("Starting advertiser")
        self.running = True
        ad_thread = Thread(target=self._advertiser, name=f"{self.name}-ad", daemon=True)
        ad_thread.start()

        print("Starting server")
        self.sserver = socketserver.ThreadingTCPServer((self.bind_ip, self.server_port), self.create_handler)
        srv_thread = Thread(target=self.sserver.serve_forever, name=f"{self.name}-srv", daemon=True, kwargs={"poll_interval" : 0.25})
        srv_thread.start()

        print("Starting listener")
        ls_thread = Thread(target=self.discover, name=f"{self.name}-ls", daemon=True)
        ls_thread.start()
    
    def shutdown(self):
        if self.sserver != None:
            self.running = False
            self.sserver.shutdown()
    
    def send(self, target: str, message: typing.Any):
        try:
            if not target in self.peers.keys():
                raise KeyError(f"Peer {target} not found")
            
            data = self.crypt.encrypt(message, self.peers[target]["public_key"])
            if self.net_encrypted:
                data = base64.urlsafe_b64encode(self.fernet.encrypt(data))
            
            with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
                # Connect to server and send data
                sock.connect((self.peers[target]["addr"].split(":")[0], int(self.peers[target]["addr"].split(":")[1])))
                sock.sendall(f"{self.network_id}|msg|{data.decode('utf-8')}|{self.pk}|END\n".encode("utf-8"))

                # Receive data from the server and shut down
                received = b""
                nr = sock.recv(1024).strip()
                while nr:
                    received += nr
                    nr = sock.recv(1024).strip()
                received += nr
            
            parts = received.decode("utf-8").split("|")
            if len(parts) != 5:
                warning(f"Malformed packet {parts} - Too many/few sections")
                return
            if parts[4] != "END":
                warning(f"Malformed packet {parts} - Packet incomplete")
                return
            if parts[0] != self.network_id:
                warning(f"Recieved response from non-network node in network {parts[0]}")
                return

            _, ptype, content, _, _ = parts

            # Check packet type
            if ptype != "rsp":
                raise ValueError(f"Packet is not a response packet: {ptype}")
            
            # Attempt to decrypt packet on network level
            if self.net_encrypted:
                try:
                    content: bytes = self.fernet.decrypt(base64.urlsafe_b64decode(content.encode("utf-8")))
                except InvalidToken:
                    raise ValueError("Bad net encryption")
            else:
                content: bytes = content.encode("utf-8")
            
            try:
                content = self.crypt.decrypt(content)
            except rsa.pkcs1.DecryptionError:
                raise ValueError("Bad node encryption")
            return content
        except:
            exception()
            return "$error"
            

class NodeHandler(socketserver.StreamRequestHandler):
    def __init__(self, node: Node, *args):
        self.node: Node = node
        super().__init__(*args)

    def handle(self) -> None:
        data: bytes = self.rfile.readline().strip()
        
        # Check packet
        parts = data.decode("utf-8").split("|")
        if len(parts) != 5:
            warning(f"Malformed packet {parts} - Too many/few sections")
            self.wfile.write(f"{self.node.network_id}|err|ERR_SECTION_COUNT||END\n".encode("utf-8"))
            return
        if parts[4] != "END":
            warning(f"Malformed packet {parts} - Packet incomplete")
            self.wfile.write(f"{self.node.network_id}|err|ERR_PACKET_INCOMPLETE||END\n".encode("utf-8"))
            return
        if parts[0] != self.node.network_id:
            self.wfile.write(f"{self.node.network_id}|err|ERR_NETWORK_INCORRECT||END\n".encode("utf-8"))
            return

        _, ptype, content, pk, _ = parts

        # Check packet type
        if not ptype in ["adv", "msg", "brd"]:
            self.wfile.write(f"{self.node.network_id}|err|ERR_PACKET_TYPE||END\n".encode("utf-8"))
            return
        
        # Attempt to decrypt packet on network level
        if self.node.net_encrypted:
            try:
                content: bytes = self.node.fernet.decrypt(base64.urlsafe_b64decode(content.encode("utf-8")))
            except InvalidToken:
                self.wfile.write(f"{self.node.network_id}|err|ERR_NET_DECRYPTION_FAILED||END\n".encode("utf-8"))
                return
        else:
            content: bytes = content.encode("utf-8")
        
        if ptype == "msg":
            # Attempt to decrypt packet on node level
            try:
                content = self.node.crypt.decrypt(content)
            except rsa.pkcs1.DecryptionError:
                self.wfile.write(f"{self.node.network_id}|err|ERR_RSA_DECRYPTION_FAILED|{self.node.pk}|END\n".encode("utf-8"))
                return
            
            result = self.node.callback(content)
            enc_result: bytes = self.node.crypt.encrypt(result, rsa.PublicKey.load_pkcs1(base64.urlsafe_b64decode(pk.encode("utf-8"))))
            if self.node.net_encrypted:
                enc_result = base64.urlsafe_b64encode(self.node.fernet.encrypt(enc_result)).decode("utf-8")
            self.wfile.write(f"{self.node.network_id}|rsp|{enc_result}|{self.node.pk}|END\n".encode("utf-8"))

        
        elif ptype == "adv":
            content = content.decode("utf-8")
            if "@" in content and ":" in content:
                name, addr = content.split("@")
                if not re.search("^((25[0-5]|2[0-4][0-9]|1[0-9][0-9]|[1-9]?[0-9])\.){3}(25[0-5]|2[0-4][0-9]|1[0-9][0-9]|[1-9]?[0-9]):[0-9]{1,5}$", addr):
                    self.wfile.write(f"{self.node.network_id}|err|ERR_BR_INVALID_IP|{self.node.pk}|END\n".encode("utf-8"))
                    return
                
                self.node.peers[name] = {
                    "name": name,
                    "addr": addr,
                    "public_key": rsa.PublicKey.load_pkcs1(base64.urlsafe_b64decode(pk.encode("utf-8")))
                }

                return
            else:
                self.wfile.write(f"{self.node.network_id}|err|ERR_BR_INVALID_FORMAT|{self.node.pk}|END\n".encode("utf-8"))
                return
        else:
            self.node.callback(content)
        

        

        


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="peernet",
    version="0.1.1",
    author="Dax Harris",
    description="Local encrypted p2p communication",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/iTecAI/pnet",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=[
        'rsa',
        'cryptography'
    ],
    keywords='p2p peer socket encryption',
    project_urls={
        'Homepage': 'https://github.com/iTecAI/pnet',
    }
)



"""
PNET - Local p2p protocol

(c) Dax Harris 2021
MIT License
"""

import rsa
import rsa.pkcs1
from cryptography.fernet import Fernet
import base64
import pickle
import random
import typing

def universal_encode(data: typing.Any) -> bytes:
    """
    Encodes any data into base64
    Yes, any data
    All the data
    :data The data
    -> The encoded data
    """
    if type(data) == str:
        _data: bytes = data.encode("utf-8")
        _fmt = b"str"
    elif type(data) == bytes:
        _data: bytes = data
        _fmt = b"raw"
    else:
        _data: bytes = pickle.dumps(data)
        _fmt = b"pkl"
    return base64.urlsafe_b64encode(_fmt+b"|"+_data)

def universal_decode(data: bytes | str) -> typing.Any:
    if type(data) == str:
        data = data.encode("utf-8")
    if b"|" in data:
        data = base64.urlsafe_b64encode(data)
    data = base64.urlsafe_b64decode(data)
    fmt, data = data.split(b"|", maxsplit=1)
    if fmt == b"raw":
        return data
    elif fmt == b"str":
        return data.decode("utf-8")
    else:
        return pickle.loads(data)
    

class Crypt:
    def __init__(self, keys: list[rsa.PublicKey | rsa.PrivateKey] | None | int = None, keygen_chance: float = 0.05):
        """
        Cryptography class for encryption
        :keys Any of [publicKey, privateKey], None (generates 256-bit keys), int n (generates n-bit keys)
        :keygen_chance The chance, per-packet, that the Fernet key will be regenerated - out of 1.0
        """
        if keys == None:
            (self.public, self.private) = rsa.newkeys(512)
        elif type(keys) == int:
            (self.public, self.private) = rsa.newkeys(keys)
        elif type(keys) == list and len(keys) == 2 and isinstance(keys[0], rsa.PublicKey) and isinstance(keys[1], rsa.PrivateKey):
            self.public = keys[0]
            self.private = keys[1]
        else:
            raise ValueError(f"Invalid keys: {keys}")
        
        self.kchance = keygen_chance
        self.key = Fernet.generate_key()
        self.fernet = Fernet(self.key)

    def encrypt(self, data: typing.Any, pk: rsa.PublicKey) -> bytes:
        """
        Encrypts data
        :data Any object. Str will be encoded, bytes will be used as-is, and anything else will be pickled.
        :pk Recipient public key
        -> base64 data
        """
        all_data = universal_encode(data)
        
        # Regenerates key randomly
        if random.random() < self.kchance:
            self.key = Fernet.generate_key()
            self.fernet = Fernet(self.key)
        
        data_encrypted = base64.urlsafe_b64encode(self.fernet.encrypt(all_data))
        enc_key = base64.urlsafe_b64encode(rsa.encrypt(self.key, pk))
        data_encoded = base64.urlsafe_b64encode(enc_key + b"|" + data_encrypted)
        return data_encoded
    
    def decrypt(self, data: str | bytes) -> typing.Any:
        """
        Decrypts data
        :data base64-encoded string or bytes
        -> Decrypted object
        """
        if type(data) == str:
            data = data.encode("utf-8")
        
        decdata = base64.urlsafe_b64decode(data)
        ekey, edata = decdata.split(b"|", maxsplit=1)
        key = rsa.decrypt(base64.urlsafe_b64decode(ekey), self.private)
        decryptor = Fernet(key)
        return universal_decode(decryptor.decrypt(base64.urlsafe_b64decode(edata)))